#ifndef PERF_QUICK_STATS_HPP
#define PERF_QUICK_STATS_HPP

#include <cmath>
#include <limits>
#include <algorithm>

namespace perf
{

/**
	A very simple stats class. Useful in cases where a full fledged boost accumulator may
	interfere with the other calculations.
	
	Refer to:
	http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#On-line_algorithm
*/
class quick_stats
{
	unsigned count_;
	double mean_, m2, min_, max_;
	
public:
	quick_stats()
	{
		reset();
	}

	/**
		Clear the data and start again.
	*/
	void reset()
	{
		count_ = 0;
		mean_ = 0;
		m2 = 0;
		min_ = std::numeric_limits<double>::max();
		max_ = std::numeric_limits<double>::min();
	}
	
	/**
		Adds a sample. All stats are immediately updated (this is an online calculation)
	*/
	void add( double x )
	{
		min_ = std::min( min_, x );
		max_ = std::max( max_, x );
		
		count_++;
		double delta = x - mean_;
		mean_ += delta/count_;
		m2 += delta*(x-mean_);
	}
	
	double mean( size_t lot_size = 1) const
	{
		return mean_ / double(lot_size);
	}
	
	double count( ) const
	{
		return count_;
	}
	
	double variance( size_t lot_size = 1 ) const
	{
		if( count_ < 2 )
			return 0;
		return m2/(count_-1) / double(lot_size * lot_size);
	}
	
	double std_dev( size_t lot_size = 1 ) const
	{
		return std::sqrt(variance(lot_size));
	}
	
	double min( size_t lot_size = 1) const
	{
		return min_ / double(lot_size);
	}
	
	double max( size_t lot_size = 1) const
	{
		return max_ / double(lot_size);
	}
};

} //eon perf

#endif