#ifndef PERF_TIMER_HPP
#define PERF_TIMER_HPP

#include "perf/quick_stats.hpp"

#include "cpp_types.hpp"

#ifdef __linux__
	#include <time.h>
#else
	#error Unsupported platform for timing
#endif

//TODO: remove this depedency, create a  library
#include <iostream>

namespace perf
{

/**	
	A  simple profiling timer to measure the time for a piece of code. It is intended
	to be used in a loop similar to the following:
	
		timer t;
		
		for( int i=0; i < loop_times; ++i )
		{
			// code to measure
			t.end();
		}

		t.report( "Report");
		
		
	For test code which is very fast and the timing overhead is significant you should
	execute the code multiple times for each timer iteration:
		
		timer t;
		
		for( int i=0; i < loop_times; ++i )
		{
			for( int j=0l j < lot_size; ++j )
			{
				// code to measure
			}
			t.end();
		}
		
		t.report( "Report", lot_size );
*/
class timer : public quick_stats
{
	typedef timespec time_t;
	time_t start_time;

	//how many recordings to exclude at the start
	size_t exclude;
public:

	timer( size_t exclude = 0)
		: exclude( exclude )
	{
		reset();
	}
	
	void reset( size_t exclude = 0)
	{
		quick_stats::reset();
		this->exclude = exclude;
		start();
	}
	
	/**
		Starts the timer (to record a timed segment). For simplicity this is automatically
		called from the constructor and the reset function. You can call start multiple
		times and only the most recent call will be relevant for the next "end" call.
	*/
	void start()
	{
		clock_gettime( CLOCK_MONOTONIC, &start_time );
	}
	
	/**
		Indicates the end of a timed segment. This automaticlly starts the next time
		segment (simply call start to override this new segment and start sometime
		later)
	*/
	void end()
	{
		time_t end;
		clock_gettime( CLOCK_MONOTONIC, &end );
	
		//add elapsed time to quick_stats
		uint64_t ns = (end.tv_sec - start_time.tv_sec)*1000000000 
			+ (end.tv_nsec - start_time.tv_nsec);

		if( exclude )
			exclude--;
		else
			add( double(ns) );
		
		//start next segment
		start_time = end;
	}

	/**
		Does a simple report of the stats to std::cout.
	*/
	void report( std::string const & label, size_t lot_size = 1 )
	{
		std::cout << label << ": "
			<< int(mean(lot_size)) << " + " << int(std_dev(lot_size)) << " ns" << std::endl
			<< "\tMin: " << int(min(lot_size)) << " Max: " << int(max(lot_size))	<< std::endl;
	}

	/**
		Prevents the compiler from removing the evaluation of the parameter (or
		at least it should if the compiler is following the standard).  This allows the
		evaluation of the parameter itself to be optimized, which is usually what
		is intended.
	*/
	template<typename T>
	void consume( T volatile item ) {}
};

} //eon perf

#endif