#include "modem.hpp"

#include <cmath>

//#define DEBUG_BITS

#ifdef DEBUG_BITS
	#include <iostream>
	#define DEBUG_LOG( x__ ) std::cout << x__
#else
	#define DEBUG_LOG( x__ )
#endif

namespace modem
{

namespace afsk
{

state::state( config const & conf )
	: conf( conf )
{
	inc_bit = float(conf.baud) / float(conf.sample_freq);

	construct_iq_filter();

	max_size_word = unsigned(std::ceil(1.0 / inc_bit) * (1/*start*/ + 8/*data*/ + 2/*stop*/ ) + cor_len);

	reset();
}

/**
	The I/Q filter is used to extract the mark and space frequency components to a segment of
	the audio data. This function simply precalculates the sin/cos values needed for the filter
	function.
*/
void state::construct_iq_filter()
{
	cor_len = unsigned( float(conf.sample_freq) / float(conf.baud) );

	iq_mark.construct( conf.mark_freq, conf.sample_freq, cor_len );
	iq_space.construct( conf.space_freq, conf.sample_freq, cor_len );
}

void state::reset()
{
	phase_bit = 0;
	blc = 0;
}

///////////////////////////////////////////////////////////
// state_writer

state_writer::state_writer( output_stream & ostream, config const & conf )
	: state( conf )
	, ostream( ostream )
{
	inc_mark = float(conf.mark_freq) / float(conf.sample_freq);
	inc_space = float(conf.space_freq) / float(conf.sample_freq);

	reset();
}

void state_writer::reset()
{
	state::reset();
	phase_wave = 0;
	bit = 0;

	//write enough header so a decoder could sync immediately at the start
	//of the stream
	modulate_bits( stop_bits, 2, true );
	DEBUG_LOG(":");
}

/**
	Writes the symbols to the audio stream. Note the term "symbols" is sometimes used
	as not all bits are data bits (they may be padding, start, or stop bits).
	
	@param direct indicates the data should be directly encoded as provided and not
		subject to any further encoding (these are non-data symbol bits)
		 (NOTE: I don't use this now, but if we wanted to do NRZ we'd have to obey
		 this and not NRZ when this is true)
		
	*NRZ I chose to use a simple encoding not the typical non-return to zero. Since this is an AFSK
	output I don't need to worry about DC level, and we are using start/stop bits which should be
	sufficient for clock sync.
	
	*PHASE In the phase part notice that we always continue the same waveform rather than switching
	between an indepent oscillator. This ensures that the output waveform is continuous and thus
	doesn't contain frequency noise.
*/
void state_writer::modulate_bits( unsigned data, unsigned len, bool direct )
{
	//safety check
	if( !len )
		return;
		
	bit = data & 1;
	data >>= 1;
	DEBUG_LOG( char( bit + '0' ) );

	while( true )
	{
		//after a bit is done writing go to the next bit
		if( phase_bit >= 1 )
		{
			blc = 0;

			phase_bit -= 1;
			if( !--len )
				break;

			//*NRZ
			//bit = (~(bit ^ data)) & 1;
			bit = data & 1;
			data >>= 1;
			DEBUG_LOG( char( bit + '0' ) );
		}

		//PERFORMANCE: use a temporary block buffer
		float amp = float(cos(phase_wave * M_PI * 2.0));
		ostream.write( &amp, 1 );

		//*PHASE
		phase_wave += bit ? inc_mark : inc_space;
		phase_bit += inc_bit;
		blc++;
	}
}

/**
	Writes a block of data to the audio stream.
*/
void state_writer::write( uint8_t const * data, uint8_t const * data_end )
{
	while( data < data_end )
	{
		//write start bit
		modulate_bits( start_bits, 1, true );
		//write the data bits
		modulate_bits( *data++, 8, false );
		//write stop bits
		modulate_bits( stop_bits, 2, true );
		
		DEBUG_LOG( "/" );
	}

	DEBUG_LOG(std::endl );
}

///////////////////////////////////////////////////////////
// state_reader
state_reader::state_reader( input_stream & istream, config const & conf )
	: state( conf )
	, istream( istream )
{
	have_sync = false;
	last_bit = false;
	last_partial = false;
	
	mark_gain = 1.0f;
	space_gain = 1.0f;

	//TODO: should be based on a calculation of theoretically expected values at a certain volume?
	//though I really can't find any test where it helps at all (other than initial sync of dead-air times)...
	hysteresis_power = 1.0;
}

/**
	Reads a single bit from the input audio data. This runs the I/Q power filter over the
	the window of data. This gives a compartive count of how often the space/mark frequency
	occurs.

	Since we slide the window cor_len times and count at each point, this will put the
	reading data about a half-bit behind in the stream compared to writing data. This is
	why the output is also pre-padded with space data first, to allow proper sync on the
	first bit.
*/
size_t state_reader::demod_bit( float const * buf, bool & bit_value, bool & bit_valid )
{
	//if we wish to skip scanning for frequencies at all samples this could be increased
	//it might harm the correctness, depending on the noise in the channel (I'm unsure)
	int const sub_sample = 1;
	blc = 0;

	int count_one = 0, count_zero = 0;
	bool sub_bit = last_partial, first_sub_bit = false;
	while( phase_bit < 1.0 )
	{
		bit_valid_pair pb = power_to_bit( buf + blc );
		if( pb.valid )
			sub_bit = pb.value;
		//TODO: if !pb.valid for many steps then we've lost the signal!
		blc += sub_sample;
		phase_bit += inc_bit * float(sub_sample);

		if( !count_one && !count_zero )
			first_sub_bit = sub_bit;

		DEBUG_LOG( (sub_bit ? '+' : '-') );
		if( sub_bit )
			count_one++;
		else
			count_zero++;
	}

	//which one wins, and is it enough
	if( count_one > count_zero )
	{
		bit_value = 1;
		bit_valid = count_one > (2*count_zero);
	}
	else
	{
		bit_value = 0;
		bit_valid = count_zero > (2*count_one);
	}

	phase_bit -= 1;

	//on switches we can do extra clock recovery
	if( bit_value != last_bit )
	{
		//the last partial may have actually belonged to this bit, thus skip back 1 in the stream
		//this helps resync and deals with an output block that is slightly slow (Test: stretch 0.98)
		if( last_partial == bit_value )
			blc--;

		//this moves us forward if it looks ilke the first sub_bit of this bit may actually belong
		//to the previous bit (Test: stretch 1.02)
		else if( last_partial == first_sub_bit )
		{
			bit_valid_pair pb = power_to_bit( buf + blc );
			if( pb.valid && pb.value == bit_value )
				blc++;
		}
	}
	
	last_bit = bit_value;
	last_partial = sub_bit;
	return blc;
}

/**
	Demodulate a sequence of bits and reconstruct the written word. The word_mask
	is used to record the correctness of each bit.
	
	FEATURE: Use 1-parity bit 1-stop bit instead of 2-stop bits. This would still help with
	clock recovery but would also allow us to recover some invalid data.
*/
size_t state_reader::demod_word( unsigned & word, unsigned & bit_errors )
{
	bit_errors = 0;
	float buf[max_size_word];

	size_t off = 0;
	word = 0;

	//look for the start bit (1)
	bool bit_value, bit_valid;

	size_t read_off = 0;
	
	//This scanning should only be done if we don't yet have clock-sync. Once
	//we have clock sync we should only retry 1-sample forward (or back?). If we continually
	//rescan we won't be able recover in presence of any invalid word.
retry_start_bit:
	istream.read( buf, max_size_word, read_off );

	bit_valid_pair bit = power_to_bit( buf );
	if( (!bit.valid || !bit.value) && (!have_sync || read_off < 4) )
	{
		read_off++;
		//give up looking if we've scanned too far (let caller determine course of action)
		if( read_off > cor_len/2 )
		{
			bit_errors = 11;
			return 0;
		}
		goto retry_start_bit;
	}

	//first bit must be start bit to be valid
	off += demod_bit( buf + off, bit_value, bit_valid );
	if( !bit_value || !bit_valid )
		bit_errors++;
	DEBUG_LOG(char(bit_value + '0'));

	for( int i=0; i < 8; ++i )
	{
		off += demod_bit( buf + off, bit_value, bit_valid );
		word |= int(bit_value) << i;

		DEBUG_LOG( (bit_valid ? char(bit_value + '0') : 'x') );
	}

	//expect stop bit
	//shift forward 1 if not a stop bit
	bit = power_to_bit( buf + off );
	if( bit.valid && bit.value )
		off++;
		
	off += demod_bit( buf + off, bit_value, bit_valid );
	if( !have_sync && (bit_value || !bit_valid) )
		bit_errors++;
	DEBUG_LOG( char(bit_value + '0') );
	
	off += demod_bit( buf + off, bit_value, bit_valid );
	if( !have_sync && (bit_value || !bit_valid) )
		bit_errors++;
	DEBUG_LOG( char(bit_value + '0') );
	DEBUG_LOG( "/" );

	have_sync = true;
	return off + read_off;
}

/**
	Read all available data from the stream.

	FEATURE: Use the word_mask to do parity recovert.
	FEATURE: report invalid bytes to the output data
*/
void state_reader::read( output_data & out )
{
	while( !istream.exhausted() )
	{
		unsigned word, bit_errors;
		size_t part = demod_word( word, bit_errors );
		if( !have_sync && bit_errors )
		{
			//keep trying to sync up
			istream.advance( 1 );
			continue;
		}
		else if( bit_errors )
		{
			bad_words++;
			//TODO: at some point we've just lost sync
		}

		istream.advance( part );
		if( bit_errors )
		{
			out.push_bad();
			DEBUG_LOG( " => BAD" << std::endl );
		}
		else
		{
			out.push_value( uint8_t(word) );
			if( word <= 32 || word >= 128 )
				DEBUG_LOG( " => <" << std::hex << word << ">" << std::dec << std::endl );
			else
				DEBUG_LOG( " => " << char(word) << std::endl );
		}
	}

	DEBUG_LOG( std::endl );
}

} //eon afsk
} //eon modem
