#ifndef MODEM_HPP
#define MODEM_HPP

#include "dsp.hpp"

#include <vector>
#include <cmath>

#include "modem_stream.hpp"

namespace modem
{

struct bit_valid_pair
{
	bool value;
	bool valid;

	bit_valid_pair( bool value, bool valid )
		: value( value ), valid( valid )
	{ }
};

/**
	Interface to write decoded data.
*/
struct output_data
{
	virtual void push_value( uint8_t value ) = 0;
	virtual void push_bad() = 0;
};

namespace afsk
{

/**
	Configuration of the data stream and encoding.
*/
struct config
{
	//the line level symbol rate (symbols per second)
	unsigned baud;
	//the audio sampling rate (Hz)
	unsigned sample_freq;
	//the mark and space frequency to use for enc/dec	
	unsigned mark_freq, space_freq;
};

/**
	Tracks the current state of the encoding or decoding.

	O = output
	I = input
*/
struct state
{
	config conf;
	
	//current phase of the symbol (I/O)
	float phase_bit;
	//phase increment for symbol (I/O)
	float inc_bit;

	//I/Q filter data
	unsigned cor_len;
	dsp::iq_filter iq_mark, iq_space;

	//bit-length-count (how many points in current bit)
	unsigned blc;

	//maximum stream size of a word + trailing filter
	unsigned max_size_word;
	
	state( config const & conf );
	void construct_iq_filter();

	void reset();

protected:
	static const unsigned start_bits = ~0;
	static const unsigned stop_bits = 0;
};

class state_reader : public state
{
public:
	state_reader( input_stream & istream, config const & conf );

	void read( output_data & out );
private:
	input_stream & istream;

	//have we successfully synced
	bool have_sync;

	//last full-bit read from the stream
	bool last_bit;
	//final sub-bit of last bit read from stream
	bool last_partial;

	//gain to apply to the space power
	double space_gain;
	double mark_gain;

	//if the power for a bit doesn't exceed these assume no change in bit value
	double hysteresis_power;
	
	size_t demod_bit( float const * buf, bool & bit_value, bool & bit_valid );
	size_t demod_word( unsigned & word, unsigned & bit_errors );

	//track how many bad words have been encountered
	unsigned bad_words;
	
	/**
		The pair returned is:
			first: The resulting bit value
			second: whether it was a strong result
	*/
	bit_valid_pair power_to_bit( float const * buf )
	{
		double mark = iq_mark.fourier_power_coefficient( buf ) * mark_gain;
		double space = iq_space.fourier_power_coefficient( buf ) * space_gain;
		return bit_valid_pair( mark > space, std::fabs(mark - space) > hysteresis_power );
	}
	
};

class state_writer : public state
{
public:
	state_writer( output_stream & ostream, config const & conf );

	void reset();
	
	void write( uint8_t const * data, uint8_t const * data_end );

private:
	output_stream & ostream;

	//current phase of the signal (O)
	float phase_wave;
	//wave phase increment for mark/space bits (O)
	float inc_mark, inc_space;
	
	//currently writing bit
	unsigned bit;
	
	void modulate_bits( unsigned data, unsigned len, bool symbol );
};

} //eon afsk
} //eon modem

#endif