#include <cmath>
#include <stdint.h>
#include <fstream>
#include <iostream>
#include <assert.h>
#include <cstring>

#include <boost/program_options.hpp>
namespace po = boost::program_options;
#include <boost/exception/diagnostic_information.hpp>

#include "dsp.hpp"
#include "mem_stream.hpp"
#include "modem.hpp"
#include "modem_codec.hpp"

uint8_t const bad_data_value = 1;

struct memory_odata : public modem::output_data
{
	std::vector<uint8_t> data;

	void push_value( uint8_t value )
	{
		data.push_back( value );
	}
	
	void push_bad()
	{
		data.push_back( bad_data_value );
	}
};

/**
	Encodes data exactly as it was provided. Length provided at creation time (only used
	for tests so this limitation is okay now).
*/
struct pass_codec : public modem::block_codec
{
	size_t len;

	pass_codec( size_t len )
		: len( len )
	{ }

	size_t get_encoded_block_len() const
	{ return len; }
	size_t get_data_block_len() const
	{ return len; }

	uint8_t decode( uint8_t const * enc_data, size_t const * erasure, size_t erasure_sz,
		uint8_t * dec_data )
	{
		memcpy( dec_data, enc_data, len );
		return 0;
	}

	void encode( uint8_t const * dec_data, uint8_t * enc_data )
	{ memcpy( enc_data, dec_data, len ); }
};

/*
	NOTE: This is very much excessively rough code. I'm coding an AFSK modem without
	completely understanding all the bits. I intend on getting something basically working
	to understand the overall concept first.
*/

using namespace std;
using namespace modem;
using namespace modem::afsk;

void dump_string( uint8_t const * data, size_t len )
{
	for( size_t i=0; i < len; ++i )
	{
		uint8_t c = data[i];
		if( c == bad_data_value )
			cout << "✖";
		else if( c <= 32 || c >= 128 )
			cout << "<" << hex << int(c) << ">" << dec;
		else
			cout << char(c);
	}
}

std::vector<uint8_t> decode( audio_vector_t const & data, block_codec & codec,
	config const & conf )
{
	memory_istream input;
	input.data  = data;
	
	state_reader read( input, conf );

	memory_odata out;

	read.read( out );

	//decode the data (look for first valid data block)
	size_t data_len = codec.get_data_block_len();
	size_t enc_len = codec.get_encoded_block_len();

	//0-pad so decoder can try right up to the end (kind of silly though to go so far)
	size_t orig_size = out.data.size();
	out.data.resize( orig_size +  enc_len, 0 );
	size_t at = 0;
	while( at < orig_size )
	{
		uint8_t block[data_len];
		uint8_t ret = codec.decode( &out.data[at], 0, 0, block );
		if( !ret )
			return std::vector<uint8_t>( block, block + data_len );

		//try next location
		++at;
	};

	//nothing decoded
	return std::vector<uint8_t>();
}

void encode( std::vector<uint8_t> data, block_codec & codec,
	config const & conf, audio_vector_t & out )
{
	//encode the data
	size_t data_len = codec.get_data_block_len();
	size_t enc_len = codec.get_encoded_block_len();
	if( data.size() > data_len )
		throw "Excessive data length"; //TODO: proper exception system
	
	uint8_t block[enc_len];
	codec.encode( &data[0], block );

	//send to the modem
	memory_ostream output;
	
	state_writer write( output, conf );
	write.write( block, block + enc_len );

	out = output.data;

	cout << "Data Length: " << data.size() << std::endl;
}

void get_test_pattern( std::vector<uint8_t> & data, int pattern )
{
	switch( pattern )
	{
		#define PATTERN( ndx_ ) \
		case ndx_: \
		{ \
			uint8_t enc_data[] = {
			
		#define PATTERN_END }; \
			data.assign( enc_data, enc_data + sizeof(enc_data) ); \
			break; \
		}
		
		PATTERN(0) 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF PATTERN_END
		PATTERN(1) 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0 PATTERN_END
		PATTERN(2) 0, 0,0,0,0,0,0,0,0 PATTERN_END
		PATTERN(3) "mort-ora-y" PATTERN_END
		PATTERN(4) "abcde/fghij/klmno/pqrst/uvwxy/z/01234/56789" PATTERN_END

		#undef PATTERN
		#undef PATTERN_END
	}
}

/**
	Applies a white noise to the sample.
*/
void apply_noise( audio_vector_t & audio, double noise )
{
	for( size_t i=0; i < audio.size(); ++i )
		audio[i] += float(((2.0 * rand() / double(RAND_MAX)) - 1.0) * noise);
}

/**
	Quantize the data into a number of integer bits.
*/
void apply_quantize( audio_vector_t & audio, int bits )
{
	float max = float(1 << (bits-2)); //-2 since there is a sign as well
	cout << "MAX: " << max << endl;
	for( size_t i=0; i < audio.size(); ++i )
		audio[i] = float( floor(audio[i] * max) / max );
}

/**
	Do a time stretch of the data. This simulates an out-of-sync clock.
*/
void apply_stretch( audio_vector_t & audio, double factor )
{
	size_t n = size_t(double(audio.size()) * factor);
	audio_vector_t out;
	out.reserve( n );

	//allow us to read past the end
	audio.push_back( 0 );

	for( size_t i=0; i < n; ++i )
	{
		double src = double(i) / factor;
		size_t ndx = size_t(src);
		
		double part = src - floor(src);
		float val = float(audio[ndx+1] * part + audio[ndx] * (1.0-part));
		out.push_back( val );
	}

	audio.swap( out );
}

void save_data( audio_vector_t & audio, std::string & filename )
{
	ofstream dump( filename.c_str(), ios::binary );
	dump.write( reinterpret_cast<char*>(&audio[0]), audio.size() * sizeof(float) );	
}

int run( int argc, char ** argv )
{
	int test_pat;
	unsigned baud;
	double noise;
	int seed;
	std::string dump_clean, dump_dirty;
	int quantize;
	double stretch;
	size_t rs_parity;
	double prefix;
	
	po::options_description desc( "Options" );
	desc.add_options()
		( "help", "show help" )
		( "test-pat", po::value( &test_pat )->default_value(-1), "use numbered test pattern" )
		( "baud", po::value( &baud )->default_value(1200), "baud rate to use" )
		( "noise", po::value( &noise )->default_value(0.0), "white noise level" )
		( "seed", po::value( &seed )->default_value(1), "random seed" )
		( "dump-clean", po::value( &dump_clean )->default_value( "/tmp/clean.out" ), "dump of clean audio data" )
		( "dump-dirty", po::value( &dump_dirty )->default_value( "/tmp/dirty.out" ), "dump of dirty audio data" )
		( "quantize", po::value( &quantize )->default_value(0), "quantize to integer bit size (0=off)" )
		( "stretch", po::value( &stretch )->default_value(1.0), "apply time stretch / bad clock" )
		( "rs-parity", po::value( &rs_parity )->default_value(0), "Use Reed Solomon with this parity level (0=don't use RS)" )
		( "prefix", po::value( &prefix  )->default_value(0), "Prefix this amount of silence to wave (seconds)" )
		;

	po::variables_map vm;
	po::store( po::command_line_parser( argc, argv ).options( desc ).run(), vm );
	po::notify( vm );

	if( vm.count( "help" ) )
	{
		cerr << desc;
		return 1;
	}

	srand( seed );
	
	std::vector<uint8_t> data;
	if( test_pat != -1 )
		get_test_pattern( data, test_pat );
		
	config conf;
	conf.baud = baud; //TODO: something not working right here...
	conf.sample_freq = 22050;
	conf.mark_freq = 1200;
	conf.space_freq = 2200;
	

	auto_ptr<block_codec> codec;
	if( rs_parity )
		codec = create_rs_codec( rs_parity, 32 );
	else
		codec.reset( new pass_codec( data.size() ) );
	
	audio_vector_t capture;

	encode( data, *codec, conf, capture );
	save_data( capture, dump_clean );

	#define LOG_AUDIO_TIME \
		cout << "Audio Time: " << (float(capture.size()) / float(conf.sample_freq)) << "s" << endl;
	LOG_AUDIO_TIME
	
	//apply disturbances
	if( prefix )
	{
		cout << "Silence Prefix: " << prefix << "s" << endl;
		size_t size = size_t(prefix * conf.sample_freq);
		audio_vector_t silence( size, 0 );
		capture.insert( capture.begin(), silence.begin(), silence.end() );
		LOG_AUDIO_TIME
	}
	
	if( noise )
	{
		cout << "Noise: " << noise << endl;
		apply_noise( capture, noise );
	}

	if( quantize )
	{
		cout << "Quantize: " << quantize << endl;
		apply_quantize( capture, quantize );
	}

	if( stretch != 1.0 )
	{
		cout << "Stretch: " << stretch << endl;
		apply_stretch( capture, stretch );
		LOG_AUDIO_TIME
	}

	save_data( capture, dump_dirty );

	std::vector<uint8_t> out = decode( capture, *codec, conf );

	cout << "IData: ";
	dump_string( &data[0], data.size() );
	cout << endl;
	
	cout << "OData: ";
	dump_string( &out[0], out.size() );
	cout << endl;

	return 0;
}

int main( int argc, char ** argv )
{
	try
	{
		return run( argc, argv );
	}
	catch( ... )
	{
		std::cerr << boost::current_exception_diagnostic_information();
		return 1;
	}
}
