#include <iostream>
#include <stdexcept>

#include "perf/timer.hpp"

/**
	A simple fixed length array class.
*/
struct array
{
	int * data;
	size_t len;
	
	array( size_t len )
		: data( new int[len] )
		, len( len )
	{ }

	~array()
	{
		delete[] data;
	}

	size_t size() const
	{
		return len;
	}
};

/**
	A length safe iterator
*/
struct safe_iterator
{
	int * at, * end;

	safe_iterator( int * at, int * end )
		: at( at )
		, end( end )
	{
	}

	bool operator!=( safe_iterator const & o )
	{
		return at != o.at;
	}

	void operator++()
	{
		at++;
	}

	int & operator*()
	{
		if( at >= end )
			throw std::out_of_range( "invalid index" );
		return *at;
	}
};

/**
	Create a common interface to create iterators so our test loop code can be
	identical between Safe/Unsafe.
*/
template<bool Safe>
struct array_access { };

//the safe access functions
template<>
struct array_access<true>
{
	typedef safe_iterator iterator;

	static iterator begin( array & a ) { return safe_iterator( a.data, a.data+ a.len ); }
	static iterator end( array & a ) { return safe_iterator( a.data + a.len, a.data + a.len ); }

	static void set( array & data, size_t at, int value )
	{
		if( at >= data.len )
			throw std::out_of_range( "invalid index" );
		data.data[at] = value;
	}
};

//the unsafe access functions
template<>
struct array_access<false>
{
	typedef int * iterator;

	static iterator begin( array & a ) { return a.data; }
	static iterator end( array & a ) { return a.data + a.len; }

	static void set( array & data, size_t at, int value )
	{
		data.data[at] = value;
	}
};

template<bool Safe>
struct fill_data_index
{
	typedef array_access<Safe> access;
	void operator()( array & data, size_t len, int value )
	{
		for( size_t i=0; i < len; ++i )
			access::set( data, i, value );
	}
};

template<bool Safe>
struct fill_data_iterator
{
	typedef array_access<Safe> access;
	void operator()( array & data, size_t len, int value )
	{
		typename access::iterator end = access::end( data );
		typename access::iterator at = access::begin( data );
		while( at != end )
		{
			*at = value;
			++at;
		}
	}
};

struct sanity_base
{
	void operator()( array & data, size_t len, int value )
	{
		for( size_t i=0; i < len; ++i )
			data.data[i] = value;
	}
};

template<typename TestFunc>
void linear_test( std::string const & msg, size_t len, TestFunc & tf )
{
	array data( len );

	std::srand(0);
	perf::timer t( 10 ); //exclude first access to not measure initial memory access

	for( size_t i=0; i < 100; ++i )
	{
		//the variable length is to trick the compiler into not optimizing and avoiding
		//the if-statment for the "safe" access. Though that it does bodes well for
		//linear safe access to memory (only random access thus suffers an overhead)
		tf(data, data.size() - (rand() % 10), int(i) );
		t.end();
	}

	t.report( msg );
}

template<bool Safe>
void perf_test( std::string const & msg )
{
	std::cout << " - - - -" << std::endl;
	fill_data_index<Safe> index;
	linear_test( msg + " Index 1MB", 1 << 20, index );
	linear_test( msg + " Index 1KB", 1 << 10, index );

	fill_data_iterator<Safe> iter;
	linear_test( msg + " Iterator 1MB", 1 << 20, iter );
	linear_test( msg + " Iterator 1KB", 1 << 10, iter );
}

int main()
{
	sanity_base sb;
	linear_test( "Plain Unsafe 1MB", 1 << 20, sb );
	
	perf_test<false>( "Unsafe" );
	perf_test<true>( "Safe" );
	return 0;
}
