#include "perf/timer.hpp"

#include <boost/program_options.hpp>
namespace po = boost::program_options;

#include <algorithm>

using namespace std;

struct benchmark
{
	typedef int value_type;
	
	//number of element sin the test set
	size_t set_size;
	//number of loop iterations to do test (for variation)
	size_t loop_size;
	//run the inner lot this number of times (required for very small sets to get proper numbers)
	size_t lot_mult;

	//data to use in the test
	scoped_array<value_type> data;

	/**
		Initializes the data to be used. This should not be done inside the actual
		test functions since the optimizer is quite clever at times. For example,
		in the run_consume if you don't initialize the data the time becomes
		constant (presumably the compiler just reads a constant rather than
		looking up the value).
	*/
	void init_data()
	{
		//always generate same random sequences to be fair (is it guaranteed that shuffle uses this?)
		srand( 1 );
		
		data.reset( new value_type[set_size] );
		for( size_t i=0; i < set_size; ++i )
			data[i] = value_type(i);
		std::random_shuffle( &data[0], &data[set_size] );
	}

	//a sorting test
	void run_sort();
	
	//an adding test
	void run_consume();
};

int main( int argc, char ** argv )
{
	benchmark bm;
	size_t set_size_k;
	
	//collect options
	po::options_description desc( "Options" );
	desc.add_options()
		( "help", "show help" )
		( "size", po::value<size_t>( &set_size_k )->required(), "Size of the set to sort (in K)" )
		( "loop-size", po::value<size_t>( &bm.loop_size )->default_value( 10000 ), "Number of loop iterations" )
		( "lot-mult", po::value<size_t>( &bm.lot_mult )->default_value( 1 ), "Multiple iterations of lot" )
		;
	
	po::variables_map vm;
	po::store( po::command_line_parser( argc, argv ).options( desc ).run(), vm );
	po::notify( vm );
	
	if( vm.count( "help" ) )
	{
		cerr << desc;
		return 1;
	}
	
	//translate K size to element size
	bm.set_size = (set_size_k * 1024) / sizeof(benchmark::value_type);
	bm.init_data();
	//bm.run_sort();
	bm.run_consume();
}

void benchmark::run_sort()
{
	perf::timer t;
	
	//run loop
	volatile value_type consume;
	for( size_t i=0; i < loop_size; ++i )
	{
		//randomize the data in each iteration
		std::random_shuffle( &data[0], &data[set_size] );
		
		//don't include shuffling in time
		t.start();
		std::sort( &data[0], &data[set_size] );
		t.end();
		//prevent removal of sort
		consume = data[set_size/2];
	}
	
	cout << "Number of Elements: " << set_size << std::endl;
	cout << "Data Size: " << (set_size * sizeof(value_type)) << std::endl;
	
	//normalize the lot_size to indicate time to sort 100 elements (1 is too small)
	size_t lot_size = size_t( double(set_size) * std::log( double(set_size) ) ) / 100;
	t.report( "Sort Time per 100", lot_size );
}

//volatile prevents the optimizer from changing the order of our loop or removing any assignments
volatile int consume;

void benchmark::run_consume()
{
	perf::timer t;
	
	//run loop
	for( size_t i=0; i < loop_size; ++i )
	{
		//step in large increments to cause cache reloading
		size_t step = 128;
		for( size_t lot=0; lot < lot_mult; ++lot )
		{
			for( size_t inc=0; inc < step; ++inc )
			{
				for( size_t j=inc; j < set_size; j += step )
					consume = data[j];
			}
		}
		t.end();
	}
	
	t.report( "Consume Time per 100", set_size/100 * lot_mult );
}
